using System;
using System.Reflection;
using HarmonyLib;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;
using System.IO;
using Newtonsoft.Json.Linq;
using Newtonsoft.Json;
using System.Runtime.CompilerServices;
using UnityEngine.Events;
using UnityEngine.UI;
using UnityEngine.EventSystems;
using System.Text;
using System.Threading.Tasks;
using Steamworks;
using Object = UnityEngine.Object;
using HMLLibrary;
using RaftModLoader;
using static Mono.Security.X509.X520;

namespace AzQuest
{
    public static class AzQuest
    {
        private static CanvasHelper _cH = null;
        public static CanvasHelper __cH
        {
            get
            {
                if (_cH is null)
                    _cH = ComponentManager<CanvasHelper>.Value;
                if (_cH.gameObject is null)
                    return null;
                return _cH;
            }
        }
        public static _QuestManager __QuestManager = null;
        public static bool hasJoinedGame = false;
        public static void InitWhenJoinGame()
        {
            _UI.Reset();
            if (TryInit_UI())
            {
                if (__QuestManager != null)
                    __QuestManager.Destroy();
                Utils.AzSaver = new AzSaver("AzQuest");
                try
                {
                    Utils.LoadConfig();
                }
                catch (Exception e)
                {
                    Console.FatalError("Utils.LoadConfig: " + e.Message);
                }
                __QuestManager = new _QuestManager();
                __QuestManager.LoadPlayerQuest();
                hasJoinedGame = true;
            }
            else
            {
                Console.Error("TryInit_UI : Error !");
                hasJoinedGame = false;
            }
        }
        public static bool TryInit_UI()
        {
            if (_UI.TryInit())
            {
                GUI_Quest.TryInit();
                GUI_QuestInventory.TryInit();
            }
            return _UI.IsInit;
        }
        public static void Destroy()
        {
            Console.Log("AzerQuest.Destroy()");
            hasJoinedGame = false;
            if (__QuestManager != null)
            {
                __QuestManager.Destroy();
                __QuestManager = null;
            }
            RecieverManager.Destroy();
            RecieverManager._QuestLocation.Clear();
            _cH = null;
            _UI.Reset();
        }
    }
    public class AzerQuest : Mod
    {
        Harmony harmonyInstance;
        string harmonyId = "raft.mod.AzQuest";
        public void Awake()
        {
            Console.modPrefix = name + " v." + version;
            harmonyInstance = new Harmony(harmonyId);
            harmonyInstance.PatchAll(Assembly.GetExecutingAssembly());
        }
        public void Start()
        {
            if (!LoadSceneManager.IsGameSceneLoaded)
            {
                Console.Log("Wait game session...");
            }
            else
            {
                AzQuest.InitWhenJoinGame();
            }
        }
        public void Update()
        {
            if (!AzQuest.hasJoinedGame)
                return;
            if (!Utils.CanWePlay)
                return;
            NetworkManager.Update();
            QuestGenerator.Update();
        }
        public void OnModUnload()
        {
            harmonyInstance.UnpatchAll(harmonyId);
            harmonyInstance = null;
            if (AzQuest.hasJoinedGame)
                AzQuest.Destroy();
        }
        public override void WorldEvent_WorldLoaded()
        {
            if (!AzQuest.hasJoinedGame)
            {
                AzQuest.InitWhenJoinGame();
            }
        }
        public override void WorldEvent_WorldUnloaded()
        {
            AzQuest.Destroy();
        }
    }
    public static class ConsoleCommands
    {
        const string HELP_CMD_P = "<color=#BD9E9E>AzerQuest command:</color> ", HELP_CMD_E = " - info: ";
        [ConsoleCommand(name: "azq.bl", docs: (HELP_CMD_P + "azq.bl " + HELP_CMD_E + "Show all items in blacklist"))]
        public static void CMD_BlackList()
        {
            Console.Log(QuestGenerator.SortBlackListToString(), Console.LogCat.Cmd);
        }
        [ConsoleCommand(name: "azq.give", docs: (HELP_CMD_P + "azq.give <%ITEM_NAME%> <AMOUNT>" + HELP_CMD_E + "Give x item to player"))]
        public static void CMD_give(string[] args)
        {
            string item_name = QuestGenerator.GameItems.FirstOrDefault(o => o.UniqueName.ToLower().Contains(args[0].ToLower())).UniqueName;
            RAPI.GetLocalPlayer().Inventory.AddItem(item_name, Convert.ToInt32(args[1]));
        }
        [ConsoleCommand(name: "azq.items", docs: (HELP_CMD_P + "azq.items" + HELP_CMD_E + "show list need/reward"))]
        public static void CMD_items()
        {
            Console.Log(QuestGenerator.SortListToString(), Console.LogCat.Cmd);
        }
        [ConsoleCommand(name: "azq.search", docs: (HELP_CMD_P + "azq.search <%ITEM_NAME%>" + HELP_CMD_E + "Search an item in QuestGenerator.GameItems"))]
        public static void CMD_searchitem(string[] args)
        {
            string mnt = " Item contains \"" + args[0] + "\" : \r\n";
            foreach (var item in QuestGenerator.GameItems.Where(o => o.UniqueName.ToLower().Contains(args[0].ToLower())))
            {
                mnt += "\r\n" + item.UniqueName + " - StackSize: " + item.settings_Inventory.StackSize + " - CraftingCategory: " + item.settings_recipe.CraftingCategory;
            }
            Console.Log(mnt, Console.LogCat.Cmd);
        }
        [ConsoleCommand(name: "azq", docs: (HELP_CMD_P + "azq <AMOUNT>" + HELP_CMD_E + "Add x quest"))]
        public static void CMD_quest(string[] args)
        {
            AzQuest.__QuestManager.AddQuest(Convert.ToInt32(args[0]));
        }
    }
    public static class Console
    {
        public static bool ShowLog = false;
        public static string modPrefix;
        private static string COLOR_END = "</b></color>] ",
            COLOR_ERROR = "[<color=#BA7878><b>",
            COLOR_FATAL_ERROR = "[<color=#CAB898><b>",
            COLOR_IMPORTANT = "[<color=#CACA98><b>",
            COLOR_CMD = "[<color=#9898CA><b>",
            COLOR_LOG = "[<color=#78BA78><b>";
        public enum LogCat
        {
            Cmd,
            Normal,
            Important,
        }
        public static void Log(string text, LogCat Category = LogCat.Normal)
        {
            switch (Category)
            {
                default:
                case LogCat.Normal:
                    if (ShowLog)
                    {
                        Debug.Log(COLOR_LOG + modPrefix + COLOR_END + text);
                    }
                    break;
                case LogCat.Important:
                    Debug.Log(COLOR_IMPORTANT + modPrefix + COLOR_END + text);
                    break;
                case LogCat.Cmd:
                    Debug.Log(COLOR_CMD + modPrefix + COLOR_END + text);
                    break;
            }
        }
        public static void Error(string text)
        {
            Debug.Log(COLOR_ERROR + modPrefix + COLOR_END + text);
        }
        public static void FatalError(string text)
        {
            Debug.Log(COLOR_FATAL_ERROR + modPrefix + text + COLOR_END);
        }
    }
    public class AzSaver
    {
        static string Dir;
        public AzSaver(string dir_name)
        {
            Dir = Path.Combine(Directory.GetCurrentDirectory(), "mods", dir_name);
            if (!Directory.Exists(Dir))
            {
                Directory.CreateDirectory(Dir);
            }
        }
        public void Save(object o, string file_name, string prefix = null)
        {
            if (prefix != null)
            {
                file_name = prefix + "_" + file_name;
            }
            try
            {
                File.WriteAllText(Path.Combine(Dir, file_name + ".json"), JsonConvert.SerializeObject(o, Formatting.Indented));
            }
            catch (Exception ex)
            {
                Console.FatalError("JSON Save error: " + ex.Message);
            }
        }
        public object Load(Type o, string file_name, string prefix = null)
        {
            if (prefix != null)
            {
                file_name = prefix + "_" + file_name;
            }
            string path = Path.Combine(Dir, file_name + ".json");
            if (File.Exists(path))
            {
                try
                {
                    return JsonConvert.DeserializeObject(File.ReadAllText(path), o);
                }
                catch (Exception ex)
                {
                    Console.FatalError("JSON Load error: " + ex.Message);
                    return null;
                }
            }
            Console.Error("JSON Load return null");
            return null;
        }
        public List<string> LoadTXT(string file_name, string prefix = null)
        {
            if (prefix != null)
            {
                file_name = prefix + "_" + file_name;
            }
            string path = Path.Combine(Dir, file_name + ".txt");
            if (!File.Exists(path))
            {
                return null;
            }
            return File.ReadAllLines(path).Where(o => !o.StartsWith("/") && o.Length > 1).ToList();
        }
        public void SaveTXT(List<string> LIST_TXT, string file_name, string prefix = null)
        {
            if (LIST_TXT is null)
                return;
            if (prefix != null)
            {
                file_name = prefix + "_" + file_name;
            }
            string path = Path.Combine(Dir, file_name + ".txt");
            File.WriteAllLines(path, LIST_TXT);
        }
    }
    public static class NetworkManager
    {
        private static float NetworkCheckInterval = 1f, NetworkNextCheck;
        public static int NetworkChannel = 99;
        public class Network_RecieverManager : Message
        {
            public Vector2 Location;
            public Network_RecieverManager(Messages e, Vector2 _Location)
            {
                Location = _Location;
            }
        }
        public static void SendToNetwork(Message o)
        {
            RAPI.SendNetworkMessage(o, NetworkChannel);
        }
        static NetworkMessage netMessage;
        public static void Update()
        {
            if (Time.realtimeSinceStartup > NetworkNextCheck)
            {
                NetworkNextCheck = Time.realtimeSinceStartup + NetworkCheckInterval;
            }
            else
            {
                return;
            }
            netMessage = RAPI.ListenForNetworkMessagesOnChannel(NetworkChannel);
            if (netMessage != null)
            {
                Message message = netMessage.message;
                if (message.Type == (Messages)9191)
                {
                    Network_RecieverManager msg = message as Network_RecieverManager;
                    RecieverManager.NetworkNewLocation(msg.Location);
                }
            }
        }
    }
    public class AzQuest_Config
    {
        public int maxItemNeed { get; set; }
        public int maxItemReward { get; set; }
        public int questAtDot { get; set; }
        public AzQuest_Config()
        {
            maxItemNeed = 30;
            maxItemReward = 50;
            questAtDot = 10;
        }
    }
    public static class Utils
    {
        public static AzSaver AzSaver;
        public static string MysteryPackage_Name = "Mystery_Package";
        public static AzQuest_Config Config;
        public static void ValidConfig()
        {
            if (Config.maxItemNeed < 1)
                Config.maxItemNeed = 1;
            if (Config.maxItemReward < 1)
                Config.maxItemReward = 1;
            if (Config.questAtDot < 1)
                Config.questAtDot = 1;
        }
        public static void LoadConfig()
        {
            var load = (AzQuest_Config)AzSaver.Load(typeof(AzQuest_Config), "config");
            if (load is null)
            {
                SaveConfig();
            }
            else
            {
                Config = load;
                ValidConfig();
            }
        }
        public static void SaveConfig()
        {
            if (Config is null)
            {
                Config = new AzQuest_Config();
            }
            AzSaver.Save(Config, "config");
        }
        public static int CountInventoryItem(List<Slot> _slots, string _UniqueName)
        {
            int count = 0;
            foreach (Slot _slot in _slots.Where(o => o.itemInstance.UniqueName == _UniqueName))
            {
                count += _slot.itemInstance.Amount;
            }
            return count;
        }
        static System.Random random = new System.Random();
        public static float RandomFloat(float min, float max)
        {
            return (float)(random.NextDouble() * (max - min) + min);
        }
        public static bool CanWePlay
        {
            get
            {
                return !Raft_Network.InMenuScene
                    && BlockCreator.GetPlacedBlocks().Count > 0
                    && (Raft_Network.IsHost || ComponentManager<Raft_Network>.Value.IsConnectedToHost);
            }
        }
    }
    public class I18NHelper
    {
        [System.Diagnostics.CodeAnalysis.SuppressMessage("Style", "IDE0060:Remove unused parameter", Justification = "Implementation unfinished. See comment in code.")]
        public static void FixI18N(GameObject gameObject, string text)
        {
            foreach (var localize in gameObject.GetComponentsInChildren<I2.Loc.Localize>())
            {
                localize.enabled = false;
                UnityEngine.Object.Destroy(localize);
            }
        }
    }
    public static class _UI
    {
        public static bool IsInit { get; private set; } = false;
        public static void Reset()
        {
            buttonPrefab = null;
            backgroundPrefab = null;
            textPrefab = null;
            dividerPrefab = null;
            IsInit = false;
        }
        private static bool CheckEverythinIsOk()
        {
            if (buttonPrefab is null)
                return false;
            if (backgroundPrefab is null)
                return false;
            if (textPrefab is null)
                return false;
            if (dividerPrefab is null)
                return false;
            return true;
        }
        public static bool TryInit()
        {
            if (buttonPrefab is null)
            {
                var inventoryResearchTable = ComponentManager<Inventory_ResearchTable>.Value;
                if (inventoryResearchTable != null)
                    buttonPrefab = inventoryResearchTable.gameObject.transform.FindChildRecursively("ResearchButton").gameObject;
            }
            var localPlayer = ComponentManager<Raft_Network>.Value.GetLocalPlayer();
            if (localPlayer != null)
            {
                var playerInventory = localPlayer.Inventory;
                if (playerInventory != null)
                {
                    if (backgroundPrefab is null)
                    {
                        backgroundPrefab = playerInventory.gameObject.transform.FindChildRecursively("BrownBackground").gameObject;
                    }
                    if (textPrefab is null)
                    {
                        textPrefab = playerInventory.gameObject.transform.FindChildRecursively("InventoryText").gameObject;
                    }
                    if (dividerPrefab is null)
                    {
                        dividerPrefab = playerInventory.gameObject.transform.FindChildRecursively("divider").gameObject;
                    }
                }
            }
            IsInit = CheckEverythinIsOk();
            Console.Log("TryInit _UI : " + (IsInit ? "OK" : "Fail"), Console.LogCat.Important);
            return IsInit;
        }
        static GameObject buttonPrefab, backgroundPrefab, textPrefab, dividerPrefab;
        public static GameObject AddButton(GameObject gameObject, string text, UnityAction onClick, float x, float y, float width = 0, float height = 0)
        {
            if (buttonPrefab != null)
            {
                var button = UnityEngine.Object.Instantiate(buttonPrefab);
                I18NHelper.FixI18N(button, text);
                button.transform.position = new Vector3(x, y, 0f);
                if (width != 0 || height != 0)
                {
                    ((RectTransform)button.transform).sizeDelta = new Vector2(width, height);
                }
                ((RectTransform)button.transform).anchorMin = new Vector2(0f, 1f);
                ((RectTransform)button.transform).anchorMax = new Vector2(0f, 1f);
                button.transform.SetParent(gameObject.transform, false);
                button.GetComponentInChildren<Text>().text = text;
                button.GetComponent<Button>().onClick.RemoveAllListeners();
                button.GetComponent<Button>().onClick = new Button.ButtonClickedEvent();
                button.GetComponent<Button>().onClick.AddListener(onClick);
                return button;
            }
            return null;
        }
        public static List<Slot> CreateSlots(GameObject gameObject, float y, int line = 1, int num = 1)
        {
            var localPlayer = ComponentManager<Raft_Network>.Value.GetLocalPlayer();
            var playerInventory = localPlayer.Inventory;
            List<Slot> slots = new List<Slot>();
            for (int j = 0; j < line; j++)
            {
                for (int i = 0; i < num; i++)
                {
                    var slot_i = CreateOneSlot(gameObject, i * 80f - 200f, y - j * 80f);
                    if (slot_i != null)
                        slots.Add(slot_i);
                }
            }
            return slots;
        }
        public static Slot CreateOneSlot(GameObject gameObject, float x, float y)
        {
            var localPlayer = ComponentManager<Raft_Network>.Value.GetLocalPlayer();
            var playerInventory = localPlayer.Inventory;
            GameObject slot = UnityEngine.Object.Instantiate(playerInventory.slotPrefab);
            slot.transform.position = new Vector3(x, y, 0f);
            slot.transform.SetParent(gameObject.transform, false);
            return slot.GetComponent<Slot>();
        }
        public static GameObject CreateBackground(GameObject gameObject, Rect _rect)
        {
            if (backgroundPrefab != null)
            {
                var background = UnityEngine.Object.Instantiate(backgroundPrefab);
                background.transform.position = new Vector3(_rect.x, _rect.y, 0f);
                ((RectTransform)background.transform).sizeDelta = new Vector2(_rect.width, _rect.height);
                ((RectTransform)background.transform).anchorMin = new Vector2(0f, 1f);
                ((RectTransform)background.transform).anchorMax = new Vector2(0f, 1f);
                background.transform.SetParent(gameObject.transform, false);
                return background;
            }
            return null;
        }
        public static GameObject AddText(GameObject gameObject, string text, float x, float y, int fontSize = 40, TextAnchor Text_Align = TextAnchor.MiddleCenter, float width = 100f)
        {
            if (textPrefab != null)
            {
                var textInstance = UnityEngine.Object.Instantiate(textPrefab);
                I18NHelper.FixI18N(textInstance, text);
                switch (Text_Align)
                {
                    case TextAnchor.MiddleLeft:
                        textInstance.transform.position = new Vector3(x, y, 0f);
                        break;
                    default:
                    case TextAnchor.MiddleCenter:
                        textInstance.transform.position = new Vector3(-width * 0.5f + x, y, 0f);
                        break;
                }
                ((RectTransform)textInstance.transform).anchorMin = new Vector2(0f, 1f);
                ((RectTransform)textInstance.transform).anchorMax = new Vector2(0f, 1f);
                ((RectTransform)textInstance.transform).sizeDelta = new Vector2(width, 100f);
                textInstance.transform.SetParent(gameObject.transform, false);
                textInstance.GetComponent<Text>().text = text;
                textInstance.GetComponent<Text>().alignment = Text_Align;
                textInstance.GetComponent<Text>().fontSize = fontSize;
                textInstance.GetComponent<Text>().resizeTextMaxSize = fontSize;
                textInstance.GetComponent<Text>().resizeTextMinSize = Math.Min(textInstance.GetComponent<Text>().resizeTextMinSize, fontSize);
                return textInstance;
            }
            return null;
        }
        public static void AddDivider(GameObject gameObject, float y)
        {
            if (dividerPrefab != null)
            {
                var divider = UnityEngine.Object.Instantiate(dividerPrefab);
                divider.transform.position = new Vector3(0f, y, 0f);
                ((RectTransform)divider.transform).sizeDelta = new Vector2(380f, ((RectTransform)divider.transform).sizeDelta.y);
                ((RectTransform)divider.transform).anchorMin = new Vector2(0f, 1f);
                ((RectTransform)divider.transform).anchorMax = new Vector2(0f, 1f);
                divider.transform.SetParent(gameObject.transform, false);
            }
        }
    }
    class GUI_Quest_Harmony
    {
        [HarmonyPatch(typeof(CanvasHelper), "OpenMenu")]
        internal class OpenMenuPatch
        {
            [HarmonyPrefix]
            private static void Prefix(CanvasHelper __instance, MenuType menuType, bool force)
            {
                if (menuType == MenuType.Inventory)
                {
                    GUI_Quest.OnInventoryShow();
                }
            }
        }
        [HarmonyPatch(typeof(CanvasHelper), "CloseMenu")]
        internal class CloseMenuPatch
        {
            [HarmonyPrefix]
            private static void Prefix(CanvasHelper __instance, MenuType menuType)
            {
                if (menuType == MenuType.Inventory)
                {
                    GUI_QuestInventory.Close();
                    GUI_Quest.OnInventoryHide();
                }
            }
        }
    }
    static class GUI_Quest
    {
        public static bool IsInit { get; private set; } = false;
        static GameObject QuestHUD;
        public static void Destroy()
        {
            Object.DestroyImmediate(QuestHUD);
            QuestHUD = null;
            IsInit = false;
        }
        public static bool TryInit()
        {
            IsInit = _TryInit();
            Console.Log("TryInit QuestHUD : " + (IsInit ? "OK" : "Fail"), Console.LogCat.Important);
            return IsInit;
        }
        private static bool _TryInit()
        {
            Console.Log("Try to init QuestHUD...");
            if (AzQuest.__cH is null)
            {
                Console.Error("__cH is null");
                return false;
            }
            if (!_UI.IsInit)
            {
                Console.Error("_UI is null");
                return false;
            }
            if (QuestHUD != null)
            {
                Destroy();
            }
            if (QuestHUD == null)
            {
                var localPlayer = ComponentManager<Raft_Network>.Value.GetLocalPlayer();
                if (localPlayer is null)
                {
                    Console.Error("localPlayer is null");
                    return false;
                }
                var playerInventory = localPlayer.Inventory;
                if (playerInventory is null)
                {
                    Console.Error("playerInventory is null");
                    return false;
                }
                QuestHUD = new GameObject("QuestHUD", typeof(RectTransform), typeof(CanvasRenderer));
                var QuestHUDTransform = QuestHUD.transform as RectTransform;
                float width = ((RectTransform)playerInventory.gameObject.transform).sizeDelta.x,
                    height = 50f, padding_top = 20f, btn_margin = 10f;
                QuestHUDTransform.SetPositionAndRotation(new Vector3(0f, playerInventory.gameObject.transform.localPosition.y + height + padding_top, 0f), Quaternion.identity);
                QuestHUDTransform.SetParent(playerInventory.gameObject.transform.parent, false);
                QuestHUDTransform.sizeDelta = new Vector2(0f, 0f);
                QuestHUDTransform.localScale = new Vector3(0.9f, 0.9f, 0.9f);
                QuestHUDTransform.anchorMax = new Vector2(0.5f, 0.5f);
                _UI.CreateBackground(QuestHUD, new Rect(0f, 0f, width + 20f, height)).transform.SetAsFirstSibling();
                GameObject btn = _UI.AddButton(QuestHUD, "Show QUest", BTN_ShowQuest, 0f, -(height * 0.5f) - btn_margin, width - (btn_margin * 2f), height - btn_margin);
                if (btn != null)
                {
                    _BTN_ShowQuest_Text = btn.GetComponentInChildren<Text>();
                }
            }
            QuestHUD.SetActiveSafe(false);
            return true;
        }
        private static Text _BTN_ShowQuest_Text;
        public static void BTN_ShowQuest()
        {
            GUI_QuestInventory.Open();
        }
        public static void OnInventoryShow()
        {
            bool canShow = CanShow();
            if (canShow)
            {
                if (AzQuest.__QuestManager.ActualQuest is null)
                {
                    if (AzQuest.__QuestManager._Count_QuestAvaible > 0 || AzQuest.__QuestManager.GetPackageCount() > 0)
                    {
                        if (AzQuest.__QuestManager._Count_QuestAvaible <= 0)
                        {
                            canShow = true;
                        }
                        else
                        {
                            if (!AzQuest.__QuestManager.GetNewQuest())
                            {
                                canShow = false;
                            }
                        }
                    }
                    else
                    {
                        canShow = false;
                    }
                }
            }
            if (canShow)
            {
                Show();
            }
        }
        public static void UpdateText()
        {
            if (_BTN_ShowQuest_Text != null)
                _BTN_ShowQuest_Text.text = (AzQuest.__QuestManager.GetVirtualQuestCount()) + " QUEST";
        }
        public static void OnInventoryHide()
        {
            Hide();
        }
        private static bool CanShow()
        {
            if (!_UI.IsInit)
                return false;
            if (!Utils.CanWePlay)
                return false;
            if (!IsInit)
            {
                if (!TryInit())
                {
                    return false;
                }
            }
            if (AzQuest.__QuestManager is null)
            {
                Console.Error("AzQuest.__QuestManager is null");
                return false;
            }
            return true;
        }
        public static void Show()
        {
            if (CanShow())
            {
                UpdateText();
                QuestHUD.SetActiveSafe(true);
            }
        }
        private static bool CanHide()
        {
            if (QuestHUD is null)
            {
                return false;
            }
            return true;
        }
        public static void Hide()
        {
            if (CanHide())
                QuestHUD.SetActiveSafe(false);
        }
    }
    public class QuestInventory : Inventory
    {
        protected override void Awake()
        {
            var playerInventory = ComponentManager<PlayerInventory>.Value;
            this.slotPrefab = playerInventory.slotPrefab;
            this.invRectTransform = GetComponent<RectTransform>();
            this.hoverTransform = (RectTransform)Instantiate(playerInventory.hoverTransform.gameObject).transform;
            this.darkenedTransform = (RectTransform)Instantiate(playerInventory.darkenedTransform.gameObject).transform;
            this.draggingImage = Instantiate(playerInventory.draggingImage.gameObject).GetComponent<Image>();
            this.draggingText = this.draggingImage.GetComponentInChildren<Text>();
            this.gridLayoutGroup = playerInventory.gridLayoutGroup;
            this.hoverTransform.SetParent(gameObject.transform, false);
            this.darkenedTransform.SetParent(gameObject.transform, false);
            this.draggingImage.transform.SetParent(gameObject.transform, false);
            Vector2 cellSize = playerInventory.gridLayoutGroup.cellSize;
            foreach (var slot in GetComponentsInChildren<Slot>())
            {
                slot.InitializeEventListeners(this);
                cellSize = slot.rectTransform.sizeDelta;
            }
            this.hoverTransform.sizeDelta = cellSize;
            this.darkenedTransform.sizeDelta = cellSize;
            this.draggingImage.rectTransform.sizeDelta = playerInventory.gridLayoutGroup.cellSize * 0.75f;
        }
        protected override void Start()
        {
            base.Start();
            this.gameObject.SetActiveSafe(true);
        }
        protected override bool ContainsSlot(Slot slot)
        {
            return base.ContainsSlot(slot);
        }
        protected override void Update()
        {
            base.Update();
            hoverTransform.gameObject.SetActiveSafe(hoverSlot != null && ContainsSlot(hoverSlot));
            if (hoverTransform.gameObject.activeInHierarchy)
                hoverTransform.position = hoverSlot.rectTransform.position;
        }
    }
    static class GUI_QuestInventory
    {
        public static bool IsInit { get; private set; } = false;
        public static bool IsOpen { get; private set; }
        private static Text QuestInventory_SwitchText;
        private static QuestInventory _QuestInventory;
        private static List<GUI_ItemSlot> _GUI_NEED_ITEM = new List<GUI_ItemSlot>();
        private static List<GUI_ItemSlot> _GUI_REWARD_ITEM = new List<GUI_ItemSlot>();
        private static Rect InitQuestInventoryUI_BG = new Rect(0f, 0f, 420f, 600f);
        private static Vector2 InitQuestInventoryUI_POS = new Vector2(415f, 380f);
        private static GameObject QuestInventoryHUD = null;
        public static void Destroy()
        {
            _QuestInventory.Clear();
            _GUI_NEED_ITEM.Clear();
            _GUI_REWARD_ITEM.Clear();
            _QuestInventory = null;
            Object.DestroyImmediate(QuestInventoryHUD);
            QuestInventoryHUD = null;
            IsInit = false;
        }
        public static bool TryInit()
        {
            IsInit = _TryInit();
            Console.Log("TryInit QuestInventoryHUD : " + (IsInit ? "OK" : "Fail"), Console.LogCat.Important);
            return IsInit;
        }
        private static bool _TryInit()
        {
            Console.Log("Try to init QuestInventoryHUD...");
            if (AzQuest.__cH is null)
            {
                Console.Error("__cH is null");
                return false;
            }
            if (!_UI.IsInit)
            {
                Console.Error("_UI is null");
                return false;
            }
            if (QuestInventoryHUD != null)
            {
                Destroy();
            }
            if (QuestInventoryHUD == null)
            {
                var localPlayer = ComponentManager<Raft_Network>.Value.GetLocalPlayer();
                if (localPlayer is null)
                {
                    Console.Error("localPlayer is null");
                    return false;
                }
                var playerInventory = localPlayer.Inventory;
                if (playerInventory is null)
                {
                    Console.Error("playerInventory is null");
                    return false;
                }
                QuestInventoryHUD = new GameObject("QuestInventory", typeof(RectTransform), typeof(CanvasRenderer));
                var QITransform = QuestInventoryHUD.transform as RectTransform;
                QITransform.SetPositionAndRotation(new Vector3(InitQuestInventoryUI_POS.x, InitQuestInventoryUI_POS.y, 0f), Quaternion.identity);
                QITransform.SetParent(playerInventory.transform.parent, false);
                QITransform.sizeDelta = new Vector2(0f, 0f);
                QITransform.localScale = new Vector3(0.9f, 0.9f, 0.9f);
                QITransform.anchorMax = new Vector2(0.5f, 0.5f);
                _UI.CreateBackground(QuestInventoryHUD, InitQuestInventoryUI_BG).transform.SetAsFirstSibling();
                _UI.AddText(QuestInventoryHUD, "Quest", 0f, -30f, 30, TextAnchor.MiddleCenter, InitQuestInventoryUI_BG.width);
                float y_i = -85f;
                _UI.AddDivider(QuestInventoryHUD, y_i);
                _UI.AddText(QuestInventoryHUD, "Demand", -(InitQuestInventoryUI_BG.width * 0.5f) + 20f, y_i + 20f, 20, TextAnchor.MiddleLeft, 200f);
                _GUI_NEED_ITEM = InitListSlot(20f, y_i - 20f);
                y_i = -340f;
                _UI.AddDivider(QuestInventoryHUD, y_i);
                _UI.AddText(QuestInventoryHUD, "Obtain", -(InitQuestInventoryUI_BG.width * 0.5f) + 20f, y_i + 20f, 20, TextAnchor.MiddleLeft, 200f);
                _GUI_REWARD_ITEM = InitListSlot(20f, y_i - 20f);
                QuestInventory_SwitchText = _UI.AddText(QuestInventoryHUD, "0 / 0", 0f, -InitQuestInventoryUI_BG.height + 30f, 20, TextAnchor.MiddleCenter).GetComponent<Text>();
                _BTN_Accept = _UI.AddButton(QuestInventoryHUD, "Accept", BTN_Accept, -120f, -InitQuestInventoryUI_BG.height + 30f);
                _BTN_Accept.SetActiveSafe(false);
                _BTN_Valid = _UI.AddButton(QuestInventoryHUD, "Valid", BTN_Valid, -120f, -InitQuestInventoryUI_BG.height + 30f);
                _BTN_Valid_Text = _BTN_Valid.GetComponentInChildren<Text>();
                _BTN_Valid.SetActiveSafe(false);
                _BTN_Cancel = _UI.AddButton(QuestInventoryHUD, "Cancel", BTN_Cancel, 120f, -InitQuestInventoryUI_BG.height + 30f);
                _BTN_Cancel_Text = _BTN_Cancel.GetComponentInChildren<Text>();
                _QuestInventory = QuestInventoryHUD.AddComponent<QuestInventory>();
            }
            QuestInventoryHUD.SetActiveSafe(false);
            return true;
        }
        private static List<GUI_ItemSlot> InitListSlot(float x, float y)
        {
            List<GUI_ItemSlot> lst = new List<GUI_ItemSlot>();
            for (int i = 0; i < 4; i++)
            {
                if (i == 2)
                {
                    x += 200f;
                    y += i * 90f;
                }
                lst.Add(Create_ItemSlot(x, y, i));
            }
            return lst;
        }
        public static void Open()
        {
            if (AzQuest.__cH is null)
            {
                Console.Error("__cH is null");
                return;
            }
            if (AzQuest.__QuestManager is null)
            {
                Console.Error("__QuestManager is null");
                return;
            }
            if (IsOpen)
                return;
            if (!IsInit)
            {
                if (!TryInit())
                {
                    return;
                }
            }
            if (AzQuest.__QuestManager.ActualQuest is null)
            {
                if (!AzQuest.__QuestManager.GetNewQuest())
                {
                    return;
                }
            }
            if (!AzQuest.__cH.OpenMenuCloseOther(MenuType.Inventory, false))
                return;
            if (AzQuest.__QuestManager.ActualQuest.recipe != null)
            {
                try
                {
                    UpdateItem(AzQuest.__QuestManager.ActualQuest.recipe);
                }
                catch (UnityException a)
                {
                    Console.FatalError(a.Message);
                }
            }
            UpdateSwitchText();
            ApplyNewBTN();
            QuestInventoryHUD.SetActiveSafe(true);
            IsOpen = true;
        }
        public static void Close()
        {
            if (!IsOpen)
                return;
            if (!IsInit)
                return;
            QuestInventoryHUD.SetActiveSafe(false);
            IsOpen = false;
        }
        private static GameObject _BTN_Accept, _BTN_Valid, _BTN_Cancel;
        private static Text _BTN_Cancel_Text, _BTN_Valid_Text;
        private static GUI_ItemSlot Create_ItemSlot(float x, float y, int i)
        {
            GUI_ItemSlot init_item_gui = new GUI_ItemSlot();
            init_item_gui._slot = _UI.CreateOneSlot(QuestInventoryHUD, -(InitQuestInventoryUI_BG.width * 0.5f) + x, y - (i * 90f));
            init_item_gui._slot.locked = true;
            init_item_gui._slot.InitializeEventListeners(_QuestInventory);
            init_item_gui._slot_text = _UI.AddText(QuestInventoryHUD, "SLOT_" + i + "_TEXT", -(InitQuestInventoryUI_BG.width * 0.5f) + x + 80f, (y - 40f) - (i * 90f), 30, TextAnchor.MiddleLeft).GetComponent<Text>();
            return init_item_gui;
        }
        public static void UpdateSwitchText(string set_text = null)
        {
            if (QuestInventory_SwitchText is null)
            {
                return;
            }
            if (set_text != null)
            {
                QuestInventory_SwitchText.text = set_text;
                return;
            }
            if (AzQuest.__QuestManager.ActualQuest is null)
            {
                QuestInventory_SwitchText.text = "QUEST NOT FOUND";
                return;
            }
            switch (AzQuest.__QuestManager.ActualQuest.state)
            {
                case _QuestManager.QuestState.None:
                    QuestInventory_SwitchText.text = "Switch : " + AzQuest.__QuestManager.GetVirtualQuestCount();
                    break;
                case _QuestManager.QuestState.Accept:
                    QuestInventory_SwitchText.text = "Accepted !";
                    break;
            }
        }
        public static void ApplyNewBTN()
        {
            if (AzQuest.__QuestManager.ActualQuest is null)
            {
                _BTN_Accept.SetActiveSafe(true);
                _BTN_Valid.SetActiveSafe(false);
                _BTN_Cancel_Text.text = "Cancel";
                return;
            }
            switch (AzQuest.__QuestManager.ActualQuest.state)
            {
                case _QuestManager.QuestState.None:
                    _BTN_Accept.SetActiveSafe(true);
                    _BTN_Valid.SetActiveSafe(false);
                    _BTN_Valid_Text.text = "Valid";
                    _BTN_Cancel_Text.text = "Cancel";
                    if (AzQuest.__QuestManager.GetVirtualQuestCount() == 0)
                    {
                        _BTN_Cancel_Text.text = "Close";
                    }
                    else
                    {
                        _BTN_Cancel_Text.text = "Next";
                    }
                    break;
                case _QuestManager.QuestState.Accept:
                    _BTN_Accept.SetActiveSafe(false);
                    _BTN_Valid.SetActiveSafe(true);
                    _BTN_Valid_Text.text = "Valid";
                    _BTN_Cancel_Text.text = "Cancel";
                    break;
                case _QuestManager.QuestState.Cancel:
                    _BTN_Valid.SetActiveSafe(true);
                    _BTN_Valid_Text.text = "Yes";
                    QuestInventory_SwitchText.text = "Sure?";
                    _BTN_Cancel_Text.text = "No";
                    break;
            }
            UpdateSwitchText();
        }
        private static void BTN_Accept()
        {
            if (AzQuest.__QuestManager.ActualQuest is null)
            {
                Console.Error("__QuestManager.ActualQuest is null");
                return;
            }
            AzQuest.__QuestManager.ActualQuest.state = _QuestManager.QuestState.Accept;
            ApplyNewBTN();
        }
        private static void BTN_Valid()
        {
            if (AzQuest.__QuestManager.ActualQuest is null)
            {
                Console.Error("__QuestManager.ActualQuest is null");
                return;
            }
            if (AzQuest.__QuestManager.ActualQuest.state == _QuestManager.QuestState.Cancel)
            {
                BTN_Cancel(true);
                return;
            }
            if (!AzQuest.__QuestManager.ValidQuest(ComponentManager<Raft_Network>.Value.GetLocalPlayer().Inventory))
            {
                Console.Log("Quest validator failed !");
                UpdateSwitchText("Fail !");
            }
            else
            {
                Console.Log("Quest complete !");
                Close();
                ApplyNewBTN();
            }
        }
        private static void BTN_Cancel(bool forceEnd = false)
        {
            if (forceEnd)
            {
                AzQuest.__QuestManager.EndOrCancelQuest(forceEnd);
                Close();
                Open();
            }
            else
            {
                BTN_Cancel();
            }
        }
        private static void BTN_Cancel()
        {
            if (AzQuest.__QuestManager.ActualQuest is null)
            {
                return;
            }
            switch (AzQuest.__QuestManager.ActualQuest.state)
            {
                case _QuestManager.QuestState.Cancel:
                    AzQuest.__QuestManager.ActualQuest.state = _QuestManager.QuestState.Accept;
                    ApplyNewBTN();
                    UpdateSwitchText();
                    break;
                case _QuestManager.QuestState.Accept:
                    AzQuest.__QuestManager.EndOrCancelQuest();
                    ApplyNewBTN();
                    UpdateSwitchText();
                    break;
                default:
                case _QuestManager.QuestState.None:
                    Close();
                    if (AzQuest.__QuestManager.GetVirtualQuestCount() > 0)
                    {
                        if (AzQuest.__QuestManager.GetNewQuest())
                        {
                            Open();
                        }
                    }
                    else
                    {
                        AzQuest.__QuestManager.EndOrCancelQuest();
                        GUI_Quest.Hide();
                    }
                    break;
            }
        }
        private static Color TextColor_Red = new Color(0.8f, 0.5f, 0.5f), TextColor_Green = new Color(0.5f, 0.8f, 0.5f);
        static void UpdateItem(QuestRecipe Recipe)
        {
            _GUI_NEED_ITEM.ForEach(o => o.Hide());
            _GUI_REWARD_ITEM.ForEach(o => o.Hide());
            int _count = 0;
            foreach (QuestRecipeItem recipe in Recipe.Need)
            {
                if (_GUI_NEED_ITEM[_count] is null)
                {
                    Console.Error("_GUI_NEED_ITEM[_count] is null");
                    return;
                }
                _GUI_NEED_ITEM[_count]._slot.sliderComponent = null;
                _GUI_NEED_ITEM[_count]._slot.SetItem(QuestGenerator.GameItems.Single(o => o.UniqueName == recipe.UniqueName), 1);
                var localPlayer = ComponentManager<Raft_Network>.Value.GetLocalPlayer();
                int count_item = Utils.CountInventoryItem(localPlayer.Inventory.allSlots.Where(o => !o.IsEmpty && o.itemInstance != null).ToList(), recipe.UniqueName);
                if (count_item > recipe.Amount)
                {
                    count_item = recipe.Amount;
                }
                _GUI_NEED_ITEM[_count]._slot_text.text = count_item + " / " + recipe.Amount;
                if (count_item < recipe.Amount)
                {
                    _GUI_NEED_ITEM[_count]._slot_text.color = TextColor_Red;
                }
                else
                {
                    _GUI_NEED_ITEM[_count]._slot_text.color = TextColor_Green;
                }
                _GUI_NEED_ITEM[_count].Show();
                _count += 1;
                if (_count > _GUI_NEED_ITEM.Count)
                {
                    Console.Error("GUI_QuestInventory.UpdateItem _count > _GUI_NEED_ITEM");
                    break;
                }
            }
            _count = 0;
            foreach (QuestRecipeItem reward in Recipe.Reward)
            {
                if (_GUI_REWARD_ITEM[_count] is null)
                {
                    Console.Error("_GUI_REWARD_ITEM[_count] is null");
                    return;
                }
                _GUI_REWARD_ITEM[_count]._slot.sliderComponent = null;
                _GUI_REWARD_ITEM[_count]._slot.SetItem(QuestGenerator.GameItems.Single(o => o.UniqueName == reward.UniqueName), 1);
                _GUI_REWARD_ITEM[_count]._slot_text.text = "x " + reward.Amount;
                _GUI_REWARD_ITEM[_count].Show();
                _count += 1;
                if (_count > _GUI_REWARD_ITEM.Count)
                {
                    Console.Error("GUI_QuestInventory.UpdateItem _count > _GUI_REWARD_ITEM");
                    break;
                }
            }
        }
        class GUI_ItemSlot
        {
            public Slot _slot = null;
            public Text _slot_text = null;
            public void Hide()
            {
                if (_slot is null) return;
                _slot.gameObject.SetActiveSafe(false);
                _slot_text.gameObject.SetActiveSafe(false);
            }
            public void Show()
            {
                if (_slot is null) return;
                _slot.gameObject.SetActiveSafe(true);
                _slot_text.gameObject.SetActiveSafe(true);
            }
        }
    }
    public static class QuestGenerator
    {
        public static List<Item_Base>
            GameItems = new List<Item_Base>(),
            reward_items = new List<Item_Base>(),
            need_items = new List<Item_Base>();
        public static void Init()
        {
            Console.Log("__QuestGenerator.Init", Console.LogCat.Important);
            LoadBlackList();
            Update_Items();
        }
        static List<string> BlackListItem = new List<string>();
        public static void LoadBlackList()
        {
            List<string> load = Utils.AzSaver.LoadTXT("ItemBlackList");
            if (load is null)
            {
                BlackListItem = new List<string>() { "// You can blacklist item here (one item unique name by line)", "Hammer", "Bucket", "PlasticCup_Empty", "PlasticCup_Water", "PlasticCup_SaltWater", "PlasticBottle_Empty", "PlasticBottle_Water", "PlasticBottle_SaltWater", };
                SaveBlackList();
                BlackListItem.RemoveAt(0);
                load = new List<string>();
                load.AddRange(BlackListItem);
            }
            BlackListItem.Clear();
            foreach (string item in load)
            {
                if (!BlackListItem.Contains(item))
                    BlackListItem.Add(item);
            }
            load = null;
        }
        public static void SaveBlackList()
        {
            Utils.AzSaver.SaveTXT(BlackListItem, "ItemBlackList");
        }
        public static bool AddItemToList(string item_name, bool toReward = false, bool toNeed = false)
        {
            if (!toReward && !toNeed)
                return false;
            if (GameItems.Exists(o => o.UniqueName == item_name))
            {
                Item_Base get_item = ItemManager.GetItemByName(item_name);
                if (get_item is null)
                    return false;
                if (toReward)
                {
                    if (!reward_items.Exists(o => o.UniqueName == get_item.UniqueName))
                    {
                        reward_items.Add(get_item);
                    }
                }
                if (toNeed)
                {
                    if (!need_items.Exists(o => o.UniqueName == get_item.UniqueName))
                    {
                        need_items.Add(get_item);
                    }
                }
                Console.Log("__QuestGenerator.AddItemToList(" + get_item.UniqueName + ", " + toReward + "," + toNeed + ")");
                return true;
            }
            else
            {
                Console.Error("__QuestGenerator.AddItemToList(" + toReward + "," + toNeed + ") Item(" + item_name + ") not found in GamesItems !");
            }
            return false;
        }
        private static float ItemsCheckInterval = 300f, ItemsNextCheck;
        private static int lastItemsCheck = 0;
        public static void Update()
        {
            if (Time.realtimeSinceStartup > ItemsNextCheck)
            {
                ItemsNextCheck = Time.realtimeSinceStartup + ItemsCheckInterval;
                Console.Log("__QuestGenerator.Update()");
            }
            else
            {
                return;
            }
            if (CountLearnedItems() != lastItemsCheck)
            {
                Update_Items();
            }
        }
        public static int CountLearnedItems()
        {
            return ItemManager.GetAllItems().Count(o => o.settings_recipe.CanCraft);
        }
        public static List<Item_Base> RemoveDoubleAndBLFrom(List<Item_Base> listFrom)
        {
            List<Item_Base> ret = new List<Item_Base>();
            foreach (Item_Base item in listFrom)
            {
                if (!ret.Contains(item))
                {
                    if (!BlackListItem.Contains(item.UniqueName))
                    {
                        ret.Add(item);
                    }
                }
            }
            return ret;
        }
        private static void Update_Items()
        {
            GameItems = ItemManager.GetAllItems();
            lastItemsCheck = GameItems.Count(o => o.settings_recipe.CanCraft);
            Item_Base[] ValidItem = GameItems.Where(o => (o.settings_recipe.CanCraft)).ToArray();
            Item_Base[] for_both_list = ValidItem.Where(o => !o.settings_buildable.Placeable && (
            o.settings_recipe.CraftingCategory == CraftingCategory.Tools
            || o.settings_recipe.CraftingCategory == CraftingCategory.Resources
            || o.settings_recipe.CraftingCategory == CraftingCategory.Weapons
            || o.settings_recipe.CraftingCategory == CraftingCategory.Decorations
            || o.settings_Inventory.StackSize > 1
            )).ToArray();
            List<Item_Base> CookItem = new List<Item_Base>();
            foreach (var item in ValidItem.Where(o => o.settings_cookable.CookingResult.item != null))
            {
                if (!CookItem.Contains(item))
                    CookItem.Add(item);
                if (!CookItem.Contains(item.settings_cookable.CookingResult.item))
                    CookItem.Add(item.settings_cookable.CookingResult.item);
            }
            foreach (var item in GameItems.Where(o => (o.settings_consumeable.FoodForm == FoodForm.Solid)))
            {
                if (!CookItem.Contains(item))
                    CookItem.Add(item);
            }
            reward_items.Clear();
            need_items.Clear();
            reward_items.AddRange(for_both_list);
            reward_items.AddRange(CookItem);
            need_items.AddRange(for_both_list);
            reward_items.AddRange(ValidItem.Where(o =>
            o.settings_recipe.CraftingCategory == CraftingCategory.Other
            ));
            need_items = RemoveDoubleAndBLFrom(need_items);
            reward_items = RemoveDoubleAndBLFrom(reward_items);
            Console.Log("__QuestGenerator.Update_Items()\r\nGameItems: " + GameItems.Count + " - need_items: " + need_items.Count + " - reward_items: " + reward_items.Count + " - blacklist_items: " + BlackListItem.Count);
            for_both_list = null;
            ValidItem = null;
            CookItem = null;
        }
        public static string SortBlackListToString()
        {
            string __ret = "";
            string ret = "\r\nBlackListItem list (" + BlackListItem.Count + ") :\r\n";
            foreach (string item in BlackListItem)
            {
                ret += item + ", ";
            }
            __ret += ret.Substring(0, ret.Length - 2) + "\r\n";
            return __ret;
        }
        public static string SortListToString()
        {
            string __ret = "";
            string ret = "\r\nReward list (" + reward_items.Count + ") :\r\n";
            foreach (Item_Base item in reward_items)
            {
                ret += item.UniqueName + " - Stack:" + item.settings_Inventory.StackSize + ", ";
            }
            __ret += ret.Substring(0, ret.Length - 2) + "\r\n";
            ret = "\r\nNeed list (" + need_items.Count + ") :\r\n";
            foreach (Item_Base item in need_items)
            {
                ret += item.UniqueName + " - Stack:" + item.settings_Inventory.StackSize + ", ";
            }
            __ret += ret.Substring(0, ret.Length - 2);
            return __ret;
        }
        public static string Test(int i)
        {
            int count = 0;
            string __ret = "";
            string ret = "\r\nTest category " + (CraftingCategory)i + ": \r\n";
            foreach (Item_Base item in ItemManager.GetAllItems().Where(o => o.settings_recipe.CraftingCategory == (CraftingCategory)i))
            {
                count++;
                ret += item.UniqueName + ", ";
            }
            __ret += ret.Substring(0, ret.Length - 2);
            return __ret + "\r\n" + count + " items";
        }
        static System.Random rng = new System.Random();
        public static QuestRecipeItem MakeOneQuestRecipeItem(ref List<Item_Base> WithList, List<string> ExcludeName = null, float AmountMultiplier = 1f)
        {
            int rng_i = rng.Next(0, WithList.Count - 1);
            string item_UniqueName = WithList[rng_i].UniqueName;
            if (ExcludeName != null)
            {
                int count_try = 3;
                while (ExcludeName.Contains(item_UniqueName))
                {
                    rng_i = rng.Next(0, WithList.Count - 1);
                    item_UniqueName = WithList[rng_i].UniqueName;
                    count_try--;
                    if (count_try < 0)
                    {
                        return new QuestRecipeItem(Utils.MysteryPackage_Name, (int)Utils.RandomFloat(1, 10));
                    }
                }
            }
            int Amount = (int)((WithList[rng_i].settings_Inventory.StackSize * Utils.RandomFloat(0.5f, 2f)) * AmountMultiplier);
            if (Amount < 1)
            {
                Amount = 1;
            }
            return new QuestRecipeItem(item_UniqueName, Amount);
        }
        public static QuestRecipe RandomQuest(int nb_need = 1, int nb_reward = 1)
        {
            QuestRecipe ret = new QuestRecipe();
            List<string> ExcludeItem = new List<string>();
            int item_need_Amount = 0;
            for (int i = 0; i < nb_need; i++)
            {
                QuestRecipeItem quest_recipe = MakeOneQuestRecipeItem(ref need_items, ExcludeItem, 0.5f);
                if (quest_recipe is null)
                    break;
                if ((item_need_Amount + quest_recipe.Amount) > Utils.Config.maxItemNeed)
                {
                    int fill_Amount = Utils.Config.maxItemNeed - item_need_Amount;
                    if (fill_Amount > 0)
                    {
                        quest_recipe.Amount = fill_Amount;
                    }
                    else
                    {
                        break;
                    }
                }
                item_need_Amount += quest_recipe.Amount;
                ExcludeItem.Add(quest_recipe.UniqueName);
                ret.Need.Add(quest_recipe);
            }
            int item_reward_Amount = 0;
            for (int i = 0; i < nb_reward; i++)
            {
                QuestRecipeItem quest_recipe = MakeOneQuestRecipeItem(ref reward_items, ExcludeItem, 1f);
                if (quest_recipe is null)
                    break;
                if ((item_reward_Amount + quest_recipe.Amount) > Utils.Config.maxItemReward)
                {
                    int fill_Amount = Utils.Config.maxItemReward - item_reward_Amount;
                    if (fill_Amount > 0)
                    {
                        quest_recipe.Amount = fill_Amount;
                    }
                    else
                    {
                        break;
                    }
                }
                item_reward_Amount += quest_recipe.Amount;
                ExcludeItem.Add(quest_recipe.UniqueName);
                ret.Reward.Add(quest_recipe);
            }
            return ret;
        }
    }
    public class _QuestManager
    {
        public void Destroy()
        {
            Console.Log("Destroy RecieverManager");
            SavePlayerQuest();
        }
        public enum QuestState
        {
            None,
            Accept,
            Cancel,
        }
        public class Quest_Progress
        {
            public QuestRecipe recipe { get; set; }
            public QuestState state { get; set; }
            [JsonConstructor]
            public Quest_Progress(PlayerQuestSave QSave)
            {
                if (QSave is null)
                    return;
                if (QSave.QuestSaved.recipe is null)
                {
                    Console.Log("QSave.QuestSaved.recipe is null");
                }
                else
                {
                    recipe = QSave.QuestSaved.recipe;
                }
                state = QSave.QuestSaved.state;
            }
            public Quest_Progress(QuestRecipe _recipe)
            {
                recipe = _recipe;
                state = QuestState.None;
            }
        }
        public class PlayerQuestSave
        {
            public int QuestCount { get; set; }
            public Quest_Progress QuestSaved { get; set; }
        }
        private PlayerQuestSave _PlayerQuestSave;
        public void LoadPlayerQuest()
        {
            var load = (PlayerQuestSave)Utils.AzSaver.Load(typeof(PlayerQuestSave), "PlayerQuest", RAPI.GetLocalPlayer().steamID.ToString());
            if (load is null)
            {
                Console.Log("LoadPlayerQuest is null > set new save");
                SavePlayerQuest();
            }
            else
            {
                _Count_QuestAvaible = load.QuestCount;
                if (load.QuestSaved != null)
                {
                    ActualQuest = new Quest_Progress(load);
                }
            }
        }
        public void SavePlayerQuest()
        {
            if (_PlayerQuestSave is null)
            {
                _PlayerQuestSave = new PlayerQuestSave();
            }
            _PlayerQuestSave.QuestCount = _Count_QuestAvaible;
            _PlayerQuestSave.QuestSaved = ActualQuest;
            Utils.AzSaver.Save(_PlayerQuestSave, "PlayerQuest", RAPI.GetLocalPlayer().steamID.ToString().ToString());
        }
        public Quest_Progress ActualQuest;
        public int GetPackageCount()
        {
            var localPlayer = ComponentManager<Raft_Network>.Value.GetLocalPlayer();
            if (localPlayer is null)
                return 0;

            return Utils.CountInventoryItem(localPlayer.Inventory.allSlots.Where(o => !o.IsEmpty && o.itemInstance != null).ToList(), Utils.MysteryPackage_Name);
        }
        public int _Count_QuestAvaible = 0;
        public int GetVirtualQuestCount()
        {
            if (AzQuest.__QuestManager._Count_QuestAvaible > 100)
            {
                AzQuest.__QuestManager._Count_QuestAvaible = 100;
            }
            return (AzQuest.__QuestManager._Count_QuestAvaible + GetPackageCount());
        }
        public _QuestManager()
        {
            Console.Log("__QuestManager.Init", Console.LogCat.Important);
            QuestGenerator.Init();
        }
        public void AddQuest(int Amount = 1)
        {
            Console.Log("AddQuest(" + Amount + ")");
            _Count_QuestAvaible += Amount;
            if (ActualQuest is null)
            {
                GetNewQuest();
            }
            GUI_Quest.Show();
        }
        public bool GetNewQuest()
        {
            bool ret = true;
            if (_Count_QuestAvaible <= 0)
            {
                if (GetPackageCount() > 0)
                {
                    var localPlayer = ComponentManager<Raft_Network>.Value.GetLocalPlayer();
                    if (localPlayer is null)
                    {
                        ret = false;
                    }
                    else
                    {
                        localPlayer.Inventory.RemoveItem(Utils.MysteryPackage_Name, 1);
                    }
                }
                else
                {
                    ret = false;
                }
            }
            else
            {
                _Count_QuestAvaible--;
            }
            if (!ret) return false;
            ActualQuest = new Quest_Progress(QuestGenerator.RandomQuest(4, 4));
            return true;
        }
        public void EndOrCancelQuest(bool forceEnd = false)
        {
            if (ActualQuest is null)
            {
                return;
            }
            if (forceEnd)
            {
                ActualQuest.state = QuestState.None;
            }
            if (ActualQuest.state == QuestState.Accept)
            {
                ActualQuest.state = QuestState.Cancel;
                return;
            }
            if (ActualQuest.state == QuestState.Cancel)
            {
                ActualQuest.state = QuestState.None;
            }
            if (ActualQuest.state == QuestState.None)
            {
                if (_Count_QuestAvaible > 0 || GetPackageCount() > 0)
                {
                    GUI_Quest.UpdateText();
                    if (GetNewQuest())
                    {
                        return;
                    }
                }
                ActualQuest = null;
                GUI_Quest.Hide();
            }
        }
        public bool ValidQuest(PlayerInventory _PlayerInventory)
        {
            if (ActualQuest is null || _PlayerInventory is null)
            {
                return false;
            }
            if (ActualQuest.state != QuestState.Accept)
            {
                return false;
            }
            if (PlayerItemManager.IsBusy)
            {
                return false;
            }
            PlayerItemManager.IsBusy = true;
            List<Slot> NoEmptySlot = _PlayerInventory.allSlots.Where(o => !o.IsEmpty && o.itemInstance != null).ToList();
            bool flag = false;
            foreach (QuestRecipeItem item in ActualQuest.recipe.Need)
            {
                int count = Utils.CountInventoryItem(NoEmptySlot, item.UniqueName);
                if (count < item.Amount)
                {
                    Console.Log("Not enought " + item.UniqueName);
                    flag = true;
                }
            }
            NoEmptySlot = null;
            if (flag)
            {
                PlayerItemManager.IsBusy = false;
                return false;
            }
            foreach (QuestRecipeItem item in ActualQuest.recipe.Need)
            {
                _PlayerInventory.RemoveItem(item.UniqueName, item.Amount);
            }
            foreach (var item in ActualQuest.recipe.Reward)
            {
                Console.Log("Received " + item.UniqueName + " x" + item.Amount + " by ValidQuest()");
                _PlayerInventory.AddItem(item.UniqueName, item.Amount);
            }
            PlayerItemManager.IsBusy = false;
            EndOrCancelQuest(true);
            return true;
        }
    }
    public class QuestRecipe
    {
        public List<QuestRecipeItem> Need { get; set; }
        public List<QuestRecipeItem> Reward { get; set; }
        public QuestRecipe()
        {
            Need = new List<QuestRecipeItem>();
            Reward = new List<QuestRecipeItem>();
        }
        [JsonConstructor]
        public QuestRecipe(List<QuestRecipeItem> _Need, List<QuestRecipeItem> _Reward)
        {
            Need = _Need;
            Reward = _Reward;
        }
    }
    public class QuestRecipeItem
    {
        public string UniqueName { get; set; }
        public int Amount { get; set; }
        public QuestRecipeItem(string _UniqueName, int _Amount)
        {
            UniqueName = _UniqueName;
            Amount = _Amount;
        }
    }
    class RecieverManagerHarmony
    {
        [HarmonyPatch(typeof(Reciever), "HandleUI")]
        static class RecieverHandleUIPatch
        {
            [HarmonyPostfix]
            static void Postfix(Reciever __instance)
            {
                if (Utils.CanWePlay && AzQuest.hasJoinedGame)
                {
                    if (__instance.hasSignal)
                    {
                        try
                        {
                            RecieverManager.OnUpdateReceiver(__instance);
                        }
                        catch (UnityException uex)
                        {
                            Console.FatalError("OnUpdateReceiver: " + uex.Message);
                        }
                    }
                }
            }
        }
    }
    public static class RecieverManager
    {
        public static void NetworkNewLocation(Vector2 Location)
        {
            Console.Log("NetworkNewLocation(" + Location.ToString() + ")");
            if (_QuestLocation != null)
            {
                if (_QuestLocation.Count(o => o.Location == Location) < 1)
                {
                    _QuestLocation.Add(new QuestLocation(Location));
                }
            }
        }
        public static void RefreshSpawnLocation()
        {
            if (Raft_Network.IsHost || Raft_Network.InSinglePlayerMode)
            {
                Console.Log("RefreshSpawnLocation()");
                var raft = ComponentManager<Raft>.Value;
                Vector2 raft2DPos = new Vector2(raft.transform.position.x, raft.transform.position.z);
                _QuestLocation.ForEach(o =>
                {
                    if ((o.Location - raft2DPos).magnitude > MaxSpawnBuoyDistance)
                    {
                        o.Destroy();
                    }
                });
                _QuestLocation.RemoveAll(o => o.CanRemove);
                int _LocationInRange = _QuestLocation.Count(IsNotTooFarAway);
                if (_LocationInRange < MaxBuoyCount)
                {
                    Vector2 TrySpawnLocation;
                    for (int u = _LocationInRange; u < MaxBuoyCount; u++)
                    {
                        TrySpawnLocation = GetRandomSpawnPoint();
                        if (TrySpawnLocation != Vector2.zero)
                        {
                            if (!Raft_Network.InSinglePlayerMode)
                            {
                                Console.Log("SendToNetwork.TrySpawnLocation(" + TrySpawnLocation.ToString() + ")");
                                NetworkManager.SendToNetwork(new NetworkManager.Network_RecieverManager((Messages)9191, TrySpawnLocation));
                            }
                            NetworkNewLocation(TrySpawnLocation);
                        }
                    }
                }
            }
        }
        public static void Destroy()
        {
            if (_QuestLocation != null)
                _QuestLocation.ForEach(o => o.Destroy());
        }
        public class QuestLocation
        {
            private bool Reach = false;
            public bool CanRemove = false;
            public void Trigger()
            {
                if (Reach)
                {
                    return;
                }
                Console.Log("TriggerDot()");
                Reach = true;
                Destroy();
                OnReachPoint();
            }
            public Vector2 Location { get; set; } = Vector2.zero;
            public Dictionary<Reciever, Reciever_Dot> RecieverDots { get; } = new Dictionary<Reciever, Reciever_Dot>();
            public QuestLocation(Vector2 location)
            {
                Location = location;
            }
            public void Destroy()
            {
                DestroyDots();
                CanRemove = true;
            }
            private void DestroyDots()
            {
                Console.Log("DestroyDots()");
                foreach (var recieverDot in RecieverDots.Values)
                {
                    Object.Destroy(recieverDot.gameObject);
                }
                RecieverDots.Clear();
                Location = Vector2.zero;
            }
        }
        public static List<QuestLocation> _QuestLocation = new List<QuestLocation>();
        private static float GetAngle(Vector2 dirOne, Vector2 dirTwo)
        {
            float num = Vector2.SignedAngle(dirOne, dirTwo);
            if (num < 0)
                num = 360f + num;
            return 360f - num;
        }
        static Color DOT_QUEST_COLOR = new Color(0.7f, 0.3f, 0.1f, 0.7f);
        private static void OnReachPoint()
        {
            AzQuest.__QuestManager.AddQuest(Utils.Config.questAtDot);
        }
        static float UpdateReceiverRefreshInterval = 5f, NextUpdateReceiver;
        public static void OnUpdateReceiver(Reciever reciever)
        {
            if (Time.realtimeSinceStartup > NextUpdateReceiver)
            {
                NextUpdateReceiver = Time.realtimeSinceStartup + UpdateReceiverRefreshInterval;
                _QuestLocation.RemoveAll(o => o.CanRemove);
                if (_QuestLocation.Count < MaxBuoyCount)
                {
                    RefreshSpawnLocation();
                }
            }
            if (_QuestLocation.Count == 0)
            {
                return;
            }
            Traverse _reciever = Traverse.Create(reciever);
            Reciever_Dot dotPrefab = _reciever.Field("dotPrefab").GetValue<Reciever_Dot>();
            RectTransform dotParent = _reciever.Field("dotParent").GetValue<RectTransform>();
            Vector2 receiver2DPos = new Vector2(reciever.transform.position.x, reciever.transform.position.z);
            QuestLocation[] bufferLocations = _QuestLocation.Where(o => !o.CanRemove).ToArray();
            foreach (var questLocation in bufferLocations)
            {
                if (!questLocation.RecieverDots.ContainsKey(reciever))
                {
                    var recieverDot = Object.Instantiate(dotPrefab, dotParent);
                    recieverDot.transform.localScale = dotPrefab.transform.localScale * 0.8f;
                    recieverDot.SetTargetedByReciever(false);
                    Traverse.Create(recieverDot).Field("rect").SetValue(recieverDot.GetComponent<RectTransform>());
                    Traverse.Create(recieverDot).Field("dotImage").GetValue<Image>().color = DOT_QUEST_COLOR;
                    questLocation.RecieverDots.Add(reciever, recieverDot);
                }
                Reciever_Dot receiverDot = questLocation.RecieverDots.GetValueSafe(reciever);
                Vector2 direction = questLocation.Location - receiver2DPos;
                float distance = direction.magnitude;
                if (distance < _reciever.Field("radarNoticePointLength").GetValue<float>())
                {
                    float angle = 360f - reciever.transform.eulerAngles.y + GetAngle(Vector2.up, direction);
                    Vector3 vector3 = -new Vector3(Mathf.Sin(angle * (Mathf.PI / 180f)), Mathf.Cos(angle * (Mathf.PI / 180f)), 0f);
                    float radarLength = _reciever.Field("radarLength").GetValue<float>();
                    float radarUIWidth = _reciever.Field("radarUIWidth").GetValue<float>();
                    float relativeDistanceOnRadar = Mathf.Clamp01(distance / radarLength);
                    float radarRadius = radarUIWidth / 2f;
                    receiverDot.gameObject.SetActiveSafe(true);
                    _reciever.Field("isCurrentlyShowingRadarDot").SetValue(true);
                    receiverDot.SetLocalPosition(vector3 * (radarRadius * relativeDistanceOnRadar));
                    receiverDot.SetLengthToPoint(distance);
                    receiverDot.SetText(distance.ToString("F0"));
                    if (distance < DistanceTriggerReach)
                    {
                        questLocation.Trigger();
                    }
                }
                else
                {
                    receiverDot.gameObject.SetActiveSafe(false);
                }
            }
        }
        public static Vector2 GetRandomSpawnPoint()
        {
            var raft = ComponentManager<Raft>.Value;
            bool foundValidPosition = false;
            int tries = 0;
            Vector2 ret = new Vector2();
            while (!foundValidPosition && tries < MaxBuoySpawnTries)
            {
                Vector2 delta = Vector2.zero;
                while (delta.sqrMagnitude == 0f)
                    delta = new Vector2(Utils.RandomFloat(-1f, 1f), Utils.RandomFloat(-1f, 1f));
                var distance = Utils.RandomFloat(MinSpawnBuoyDistance, MaxSpawnBuoyDistance);
                delta = delta.normalized * distance;
                ret = new Vector2(raft.transform.position.x + delta.x, raft.transform.position.z + delta.y);
                foundValidPosition = IsValidSpawn(ret);
                tries++;
            }
            if (foundValidPosition)
            {
                return ret;
            }
            else
            {
                return Vector2.zero;
            }
        }
        private static bool IsValidSpawn(Vector2 position)
        {
            foreach (var buoyLocation in _QuestLocation)
            {
                var distance = (buoyLocation.Location - position).magnitude;
                if (distance < (BuoyCollisionOverlapRadius * 2))
                    return false;
            }
            return true;
        }
        public static bool IsNotTooFarAway(QuestLocation buoyLocation)
        {
            var raft = ComponentManager<Raft>.Value;
            Vector2 raft2DPos = new Vector2(raft.transform.position.x, raft.transform.position.z);
            float distance = (buoyLocation.Location - raft2DPos).magnitude;
            return distance <= MaxSpawnBuoyDistance;
        }
        public static float DistanceTriggerReach = 100f;
        public static int MaxBuoySpawnTries { get; } = 8;
        public static int BuoyCollisionOverlapRadius { get; } = 500;
        public static int MaxSpawnBuoyDistance { get; } = 2500;
        public static int MinSpawnBuoyDistance { get; } = 500;
        public static int MaxBuoyCount { get; } = 5;
    }
}
